# Author: Gael Varoquaux <gael.varoquaux at normalesup.org>
# Copyright (c) 2007-2020, Enthought, Inc.
# License: BSD Style.

print("!! mayavi.tools.mlab is obsolete and has been replaced by !!")
print("!! mayavi.mlab. Please update your code.                  !!")

from mayavi.mlab import *
